#include "E.h"

void MapUnmap(int start)
{
   static Window *wlist=NULL;
   Window par;
   Window rt;
   XWindowAttributes attr;
   static unsigned int num=0;
   int i;
   
   switch(start)
     {
      case 0:
	XQueryTree(disp,root.win,&rt,&par,&wlist,&num);
	if (wlist)
	  {
	     for (i=0;i<num;i++) 
	       {
		  XGetWindowAttributes(disp,wlist[i],&attr);
		  if (attr.map_state==IsUnmapped)
		    wlist[i]=0;
		  else
		    XUnmapWindow(disp,wlist[i]);
	       }
	  }
	break;
      case 1:
	if (wlist)
	  {
	     for (i=0;i<num;i++) 
	       {
		  if (wlist[i])
		    {
		       XGetWindowAttributes(disp,wlist[i],&attr);
		       if (attr.override_redirect)
			 XMapWindow(disp,wlist[i]);
		       else
			 {
			    AddToFamily(wlist[i]);
			 }
		    }
	       }
	     XFree(wlist);
	  }
	break;
      default:
	break;
     }
}

void SetupSignals()
{
   struct sigaction sa;
   
   sa.sa_handler = HandleSigChild;
   sa.sa_flags = 0;
   sigemptyset (&sa.sa_mask);
   sigaction(SIGCHLD, &sa, (struct sigaction *)0);
}

void SetupX()
{
   int test_event_base,test_error_base,test_v1,test_v2;
   int shape_event_base,shape_error_base;
   
   /* Open a connection to the diplay nominted by the DISPLAY variable */
   disp=XOpenDisplay(NULL);
   /* if cannot connect to display */
   if (!disp)
     {
	Alert("Enlightenment cannot connect to the display nominated by\n"
	      "your shell's DISPLAY environment variable. You may set this\n"
	      "variable to indicate which display name Enlightenment is to\n"
	      "connect to. It may be that you do not have an Xserver already\n"
	      "running to serve that Display connection, or that you do not\n"
	      "have permission to connect to that display. Please make sure\n"
	      "all is correct before trying again. Run an Xserver by running\n"
	      "xdm or startx first, or contact your local system\n"
	      "administrator, or Xserver vendor, or read the X, xdm and\n"
	      "startx manual pages before proceeding.\n");
	EExit(1);
     }
   /* set up an error handler for then E would normally have fatal X errors */
   XSetErrorHandler((XErrorHandler)HandleXError);
   /* set up a handler for when the X Connection goes down */
   XSetIOErrorHandler((XIOErrorHandler)HandleXIOError);
   /* Check for the Shape Extension */
   if (!XShapeQueryExtension(disp,&shape_event_base,&shape_error_base))
     {
	Alert("FATAL ERROR:\n"
	      "This Xserver does not support the Shape extension.\n"
	      "This is required for Enlightenment to run.\n"
	      "Exiting.\n");
	EExit(1);
     }
   /* check for the XTEST extension */
   if (XTestQueryExtension(disp,&test_event_base,&test_error_base,&test_v1,&test_v2))
     {
	/* make Enlightenment immune to other clients using XGrabServer */
	XTestGrabControl(disp,True);
     }   
   else
     Alert("WARNING:\n"
	   "This Xserver does not support the XTest extension.\n"
	   "This is required for Enlightenment to run properly.\n"
	   "Enlightenment will continue to run, but parts may not.\n"
	   "Work correctly.\n"
	   "Please contact your system administrator, or see the manuals\n"
	   "For your XServer to find out how to enable the XTest\n"
	   "Extension\n");
   /* record the event base for shape change events */
   event_base_shape=shape_event_base;
   /* initialise imlib */
   id=Imlib_init(disp);
   if (!id)
     {
	Alert("FATAL ERROR:\n"
	      "Enlightenment is unable to initialise Imlib.\n"
	      "Exiting.\n");
	EExit(1);
     }
   fd=Fnlib_init(id);
   if (!fd)
     {
	Alert("FATAL ERROR:\n"
	      "Enlightenment is unable to initialise Fnlib.\n"
	      "Exiting.\n");
	EExit(1);
     }
   root.win=id->x.root;
   root.vis=Imlib_get_visual(id);
   root.depth=id->x.depth;
   root.cmap=Imlib_get_colormap(id);
   root.scr=DefaultScreen(disp);
   root.w=DisplayWidth(disp,root.scr);
   root.h=DisplayHeight(disp,root.scr);
   /* warn, if necessary about visual problems */
   if (DefaultVisual(disp,root.scr)!=root.vis)
     {
	Alert("WARNING:\n"
	      "The Visual used for the Root Window (Desktop Background) does\n"
	      "not match the visual that Imlib has chosen to use (which\n"
	      "will the the visual with the greatest bit depth / greatest\n"
	      "number of colours). This may result in backgrounds set by\n"
	      "Enlightenment either not working, or looking strange.\n"
	      "You can fix this by running your Xserver with its default\n"
	      "Visual being the one with the highest bit-depth. Please\n"
	      "refer to your Xserver manuals, or contact your system\n"
	      "administrator or Xserver vendor.\n");
     }
   /* warn, if necessary about lack fo multi-head support */
   if (ScreenCount(disp)>1)
     {
	Alert("WARNING:\n"
	      "Your Xserver supports a multi-headed configuration with\n"
	      "multiple screens attached to the one display. Enlightenment\n"
	      "does not currently support multi-headed setups, and as a\n"
	      "result will only manage the first screen (this one).\n");
     }
   /* warn, if necessary about X version problems */
   if (ProtocolVersion(disp)!=11)
     {
	Alert("WARNING:\n"
	      "This is not an X11 Xserver. It infact talks the X%i protocol.\n"
	      "This may mean Enlightenment will either not function, or\n"
	      "function incorrectly. If it is later than X11, then your\n"
	      "server is one the author(s) of Enlightenment neither have\n"
	      "access to, nor have heard of.\n",ProtocolVersion(disp));
     }
    /* sleect all the root window events to start managing */
   XSelectInput(disp,root.win,
    		SubstructureNotifyMask|ButtonPressMask|ButtonReleaseMask|
    		EnterWindowMask|LeaveWindowMask|ButtonMotionMask|
    		PropertyChangeMask|SubstructureRedirectMask|KeyPressMask|
    		KeyReleaseMask|PointerMotionMask);
   XSync(disp,False);
   ICCCM_SetIconSizes();
   MWM_SetInfo();
   initFunctionArray();
   mode.mode=MODE_NONE;
   mode.ewin=NULL;
   mode.noewin=0;
   mode.px=0;mode.py=0;
   mode.x=0;mode.y=0;
   mode.focusmode=FOCUS_SLOPPY;
   mode.focuswin=NULL;
   mode.mouse_over_win=NULL;
   mode.click_focus_grabbed=0;
   mode.movemode=0;
   mode.resizemode=1;
   mode.slidemode=0;
   mode.cleanupslide=1;
   mode.mapslide=1;
   mode.slidespeedmap=4000;
   mode.slidespeedcleanup=6000;
   mode.doingslide=0;
   InitDesktopBgs();
   GotoDesktop(0);
}

void SetupDirs()
{
   char *home;
   char s[1024],ss[1024];
   
   home=homedir(getuid());
   sprintf(s,"%s/.enlightenment",home);
   if (exists(s))
     {
	if (isfile(s))
	  {
	     Alert("%s is a file not a directory. %s will be renamed %s.old and\n"
		   "a %s direcotry will be created.\n",s,s,s);
	     sprintf(ss,"%s/.enlightenment.old",home);
	     mv(s,ss);
	     md(s);
	  }
     }
   else 
     {
	md(s);
	Alert("This is just a friendly message that the following directory:\n"
	      "%s\n"
	      "has been created by Enlightenment. You are free to chnage the\n"
	      "permissions on it as you please, as long as you are able to\n"
	      "read and write to the directory's contents, which Enlightenemnt\n"
	      "will fill with user-dependant information as time goes on.\n",s);
     }
   Efree(home);
}
